const { Transform } = require('stream');
const { checkRule } = require('./lib');

/**
 * Create a Transform stream which will maintain a buffer with data received from a Readable stream and write data when the buffer can be matched against the regex. It will push the whole match object (or objects when the g flag is used) returned by `/regex/.exec(buffer)`.
 * @param {RegExp} regex Regular expression to execute
 */
               function createRegexTransformStream(regex) {
  let buffer = ''

  const ts = new Transform({
    transform(chunk, encoding, next) {
      let lastMatch
      let match
      buffer += chunk.toString()

      // thx stream-snitch
      // https://github.com/dmotz/stream-snitch/blob/master/index.js#L52
      // eslint-disable-next-line no-cond-assign
      while (match = regex.exec(buffer)) {
        ts.push(match)
        lastMatch = match
        if (!regex.global) break
      }
      if (lastMatch) {
        buffer = buffer.slice(lastMatch.index + lastMatch[0].length)
      }
      next()
    },
    objectMode: true,
  })

  return ts
}

       class Replaceable extends Transform {
  /**
   * Replaceable class that extends Transform and pushes data when it's done replacing each incoming chunk. If the replacement is passed as a function, it will work in the same way as the replacer for `string.replace` method (https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String/replace), taking the `match` as the first argument, and matched `p1`, `p2`, _etc_ parameters as following arguments. The replacer can also be an async function.
   * @constructor
   * @param {Rule|Rule[]} rules A single replacement rule, or multiple rules.
   * @example
   *
   * // markdown __ to html emphasise implementation
   * const stream = replaceStream({
   *  re: /__(\S+)__/g,
   *  replacement(match, p1) {
   *    return `<em>${p1}</em>`
   *  },
   * })
   */
  constructor(rules) {
    super()
    const re = Array.isArray(rules) ? rules : [rules]
    const fr = re.filter(checkRule)
    this.rules = fr
  }

  async reduce(chunk) {
    /** @type {string} */
    const s = await this.rules.reduce(async (acc, { re, replacement }) => {
      /** @type {string} */
      let string = await acc
      if (typeof replacement == 'string') {
        string = string.replace(re, replacement)
      } else {
        /** @type {function} */
        const R = replacement.bind(this)
        const promises = []
        const t = string.replace(re, (match, ...args) => {
          const p = R(match, ...args)
          if (p instanceof Promise) {
            promises.push(p)
          }
          return p
        })
        if (promises.length) {
          const data = await Promise.all(promises)
          string = string.replace(re, () => data.shift())
        } else {
          string = t
        }
      }
      return string
    }, `${chunk}`)

    return s
  }
  async _transform(chunk, _, next) {
    try {
      const s = await this.reduce(chunk)
      this.push(s)
      next()
    } catch (err) {
      next(err)
    }
  }
}





const $_lib_markers = require('./lib/markers');

/* documentary types/rules.xml */
/**
 * @typedef {(match: string, ...params: string[]) => string} Replacer
 *
 * @typedef {(match: string, ...params: string[]) => Promise.<string>} AsyncReplacer
 *
 * @typedef {Object} Rule A replacement rule.
 * @prop {RegExp} re A Regular expression to match against.
 * @prop {string|Replacer|AsyncReplacer} replacement A replacement string, or a replacement string function.
 */


module.exports = createRegexTransformStream
module.exports.Replaceable = Replaceable
module.exports.makeMarkers = $_lib_markers.makeMarkers
module.exports.makeCutRule = $_lib_markers.makeCutRule
module.exports.makePasteRule = $_lib_markers.makePasteRule
//# sourceMappingURL=index.js.map